const _ = require('lodash');

const summaryTemplate = '' +
  '=============================== Coverage summary ===============================\n' +
  'Total: {{ total.percent }}% covered\n' +
  'Statements: {{ statements.percent }}% ({{ statements.covered }}/{{ statements.total }})<% if(statements.skipped > 0) { print(", " + statements.skipped + " ignored") } %>\n' +
  'Branches: {{ branches.percent }}% ({{ branches.covered }}/{{ branches.total }})<% if(branches.skipped > 0) { print(", " + branches.skipped + " ignored") } %>\n' +
  'Functions: {{ functions.percent }}% ({{ functions.covered }}/{{ functions.total }})<% if(functions.skipped > 0) { print(", " + functions.skipped + " ignored") } %>\n' +
  'Lines: {{ lines.percent }}% ({{ lines.covered }}/{{ lines.total }})<% if(lines.skipped > 0) { print(", " + lines.skipped + " ignored") } %>\n' +
  '================================================================================\n';

class SimpleCoverageConfigBuilder {

  constructor(parent) {
    // Configure properties object
    Object.defineProperty(this, 'properties', {
      value: {
        parent: parent,
        configuration: this._defaultConfiguration()
      },
      enumerable: false
    });
  }

  get _subConfigurationGroupName() {
    return 'coverage';
  }

  _defaultConfiguration() {
    return {
      enabled: true,
      reportDirectory: 'coverage',
      excludePatterns: ['**/test/**', '**/node_modules/**', '**.spec.js'],
      summaryTemplate: summaryTemplate,
      enableSummary: undefined,
      nativeConfiguration: {}
    };
  }

  clear() {
    this.properties.configuration = this._defaultConfiguration();
    return this;
  }

  enable() {
    this.properties.configuration.enabled = true;
    return this;
  }

  disable() {
    this.properties.configuration.enabled = false;
    return this;
  }

  setEnabled(value) {
    this.properties.configuration.enabled = !!value;
    return this;
  }

  reportDirectory(value) {
    if (_.isNil(value)) {
      this.properties.configuration.reportDirectory = this._defaultConfiguration().reportDirectory;
    } else if (_.isString(value)) {
      this.properties.configuration.reportDirectory = value;
    }

    return this;
  }

  excludePatterns(value) {
    if (_.isArray(value)) {
      this.properties.configuration.excludePatterns = _.filter(value, item => _.isString(item));
    } else if (_.isString(value)) {
      this.properties.configuration.excludePatterns = [value];
    } else if (_.isNil(value)) {
      this.properties.configuration.excludePatterns = this._defaultConfiguration().excludePatterns;
    }

    return this;
  }

  addExcludePatterns(value) {
    if (_.isArray(value) || _.isString(value)) {
      this.properties.configuration.excludePatterns = _.concat(this.properties.configuration.excludePatterns, value);
    }

    return this;
  }

  nativeConfiguration(value) {
    if (_.isNil(value)) {
      this.properties.configuration.nativeConfiguration = this._defaultConfiguration().nativeConfiguration;
    } else if (_.isObject(value)) {
      this.properties.configuration.nativeConfiguration = value;
    }

    return this;
  }

  summaryTemplate(value) {
    if (_.isNil(value)) {
      this.properties.configuration.summaryTemplate = this._defaultConfiguration().summaryTemplate;
    } else if (_.isString(value)) {
      this.properties.configuration.summaryTemplate = value;
    }

    return this;
  }

  enableSummary(value) {
    if (_.isNil(value)) {
      this.properties.configuration.enableSummary = this._defaultConfiguration().enableSummary;
    } else if (_.isBoolean(value)) {
      this.properties.configuration.enableSummary = value;
    } else if (_.isString(value)) {
      this.properties.configuration.enableSummary = value.toLowerCase() === 'true';
    }

    return this;
  }

  parentBuilder() {
    return this.properties.parent;
  }

  build() {
    const userConfig = {
      istanbul: {
        excludes: this.properties.configuration.excludePatterns,
        nativeConfiguration: this.properties.configuration.nativeConfiguration,
        reportDirectory: this.properties.configuration.reportDirectory
      },
      enabled: this.properties.configuration.enabled,
      summary: {
        enabled: this.properties.configuration.enableSummary,
        template: this.properties.configuration.summaryTemplate
      }
    };

    return userConfig;
  }
}

exports.SimpleCoverageConfigBuilder = module.exports.SimpleCoverageConfigBuilder = SimpleCoverageConfigBuilder;
