const _ = require('lodash'),
  Joi = require('joi'),
  pathUtil = require('path'),
  debugUtils = require('debug'),
  classUtils = require('./utils/class-utils').utils;

function lodashMergerWithIgnoreNil(objValue, srcValue) {
  if (_.isNil(srcValue)) {
    return objValue;
  }
}

const States = {
  None: 'none',
  Running: 'running',
  Ready: 'ready',
  PostExecution: 'postExecution',
  ShuttingDown: 'shuttingDown'
};

class SimpleTestLauncher {

  static get States() {
    return States;
  }

  constructor() {

    Reflect.defineProperty(this, 'logger', {
      value: {
        debug: debugUtils('ze:test:launcher:debug'),
        info: debugUtils('ze:test:launcher:info'),
        warn: debugUtils('ze:test:launcher:warn'),
        error: debugUtils('ze:test:launcher:error')
      },
      enumerable: false
    });

    Reflect.defineProperty(this, 'properties', {
      value: {
        currentState: SimpleTestLauncher.States.None,
        executionContext: {},
        processExitCode: 0,
        // Private data for istanbul
        istanbul: {
          // The istanbul configuration object
          config: undefined,
          coverageVariableName: `$$cov_${Date.now()}$$`
        },
        // Private data for jasmine
        jasmine: {
          // Contains the jasmine instance
          instance: undefined,
          lastExecution: {
            success: false
          }
        },

        // Contains all the steps
        steps: {
          preLaunch: [],
          preTestLaunch: [],
          postExecution: [],
          cleanUp: []
        },

        // Contains the configuration before analyzing
        rawInputConfiguration: {
          configurationFunc: undefined,
          fromEnvVariables: undefined
        },

        configuration: {
          project: {
            rootDirectoryPath: undefined
          },
          reports: {
            rootDirectoryPath: undefined
          },
          test: {
            type: 'unknown',
            specDirPath: undefined,
            // Help to filter spec files
            specFileFilters: [],
            jasmine: {
              features: {
                reports: {
                  enableJunitXml: true,
                  enableHtml: true
                }
              },
              lastExecution: {
                success: false
              }
            }
          },
          coverage: {
            enabled: !process.env.ZE_DISABLE_CODE_COVERAGE,
            istanbul: {
              excludes: ['**/test/**', '**/node_modules/**'],
              nativeConfiguration: undefined,
              reportDirectory: 'coverage'
            },
            summary: {
              enabled: false,
              template: ''
            }
          }
        }
      },
      enumerable: false
    });
  }

  get currentState() {
    return this.properties.currentState;
  }

  set currentState(value) {
    this.properties.currentState = value;
  }


  getConfiguration() {
    return this.properties.configuration;
  }

  getProjectRootDirPath() {
    return this.properties.configuration.project.rootDirectoryPath;
  }

  getTestType() {
    return this.properties.configuration.test.type;
  }

  getRootReportPath() {
    return this.properties.configuration.reports.rootDirectoryPath || `${this.getProjectRootDirPath()}/reports/test`;
  }

  getRootReportTestPath() {
    return `${this.getRootReportPath()}/${this.getTestType()}`;
  }

  getTestSpecDirPath() {
    if (this.properties.configuration.test.specDirPath) {
      // User specify the directory to use
      return this.properties.configuration.test.specDirPath;
    } else {
      // Generate using the standard
      return `test/${this.getTestType()}`;
    }
  }

  isCodeCoverageEnabled() {
    return this.properties.configuration.coverage.enabled;
  }

  getInputConfiguration() {
    return this.properties.inputConfiguration;
  }

  isReady() {
    return this.currentState === SimpleTestLauncher.States.Ready;
  }

  isRunning() {
    return this.currentState === SimpleTestLauncher.States.Running;
  }

  async fetchJasmineConfiguration() {
    const defaultJasmineConfiguration = {
      spec_dir: this.getTestSpecDirPath(),
      spec_files: [
        "**/*[sS]pec.js",
        '*[sS]pec.js'
      ],
      // Jasmine 3 : Require env.configure
      env: {
        random: false,
        oneFailurePerSpec: false
      }
    };

    const jasmineConfiguration = _.mergeWith({},
      defaultJasmineConfiguration,
      this.properties.configuration.test.jasmine,
      lodashMergerWithIgnoreNil
    );

    return {
      configuration: jasmineConfiguration
    };
  }

  async fetchIstanbulConfiguration() {
    const inputConfiguration = this.properties.configuration.coverage.istanbul;

    const reportsDirectory = `${this.getRootReportTestPath()}/${inputConfiguration.reportDirectory}`;

    const defaultConfiguration = {
      instrumentation: {
        root: this.getProjectRootDirPath(),
        excludes: inputConfiguration.excludes
      },
      reporting: {
        print: "summary",
        dir: reportsDirectory
      },
      reports: ['lcov'],
      dir: reportsDirectory
    };

    const istanbulConfiguration = _.mergeWith({},
      defaultConfiguration,
      inputConfiguration.nativeConfiguration,
      lodashMergerWithIgnoreNil
    );

    return {
      configuration: istanbulConfiguration
    };
  }

  async _startCoverage() {
    const __pretty_name__ = '_startCoverage';

    const istanbul = require('istanbul');

    global[this.properties.istanbul.coverageVariableName] = {};

    const {configuration: generatedConfig} = await this.fetchIstanbulConfiguration();

    const configuration = istanbul.config.loadObject(generatedConfig);
    this.properties.istanbul.config = configuration;

    const excludes = configuration.instrumentation.excludes(true);

    const matcherFor = () => {
      return new Promise((resolve, reject) => {
        istanbul.matcherFor({
          root: configuration.instrumentation.root() || process.cwd(),
          includes: configuration.instrumentation.extensions().map(ext => '**/*' + ext),
          excludes: excludes
        }, (error, matchFn) => {
          if (error) {
            // The fatal kill the process
            this.logger.error("[%s] Failed to create istanbul matcher: %s\n%s", __pretty_name__, error.message, error.stack);
            return reject(error);
          }

          // Using intrumentor with es6
          const createInstrumenter = require('istanbul-lib-instrument').createInstrumenter;

          const instrumenter = createInstrumenter({
            coverageVariable: this.properties.istanbul.coverageVariableName,
            preserveComments: true,
            esModules: true
          });
          const transformer = instrumenter.instrumentSync.bind(instrumenter);

          // Enable the require hook
          istanbul.hook.hookRequire(matchFn, transformer);
          resolve();
        });
      });
    };

    await matcherFor();
  }

  async _startTest() {
    this.properties.jasmine.instance.onComplete(passed => {
      this.properties.jasmine.lastExecution.success = !!passed;
      setImmediate(() => this._handleTestFinished());
    });

    this.logger.info("Start the execution");
    this.properties.jasmine.instance.execute();
  }

  async _handleTestFinished() {
    const __pretty_name__ = '_handleTestFinished';

    // Do the post execution
    this.currentState = SimpleTestLauncher.States.PostExecution;

    // Execute the steps post execution
    const postExecutionContext = this._createStepBaseContext();
    postExecutionContext.testSucceed = this.properties.jasmine.lastExecution.success;
    // Set the exit code
    this.logger.debug('[%s] Jasmine return success=%s', __pretty_name__, this.properties.jasmine.lastExecution.success);
    this.properties.processExitCode = this.properties.jasmine.lastExecution.success ? 0 : 1;

    try {
      await this._execSteps({
        groupName: 'postExecution',
        context: postExecutionContext,
        stopOnError: false,
        setExitCode: code => {
          if (_.isNumber(code)) {
            this.properties.processExitCode = code;
          }
        }
      });

      const cleanUpContext = this._createStepBaseContext();
      await this._execSteps({
        groupName: 'cleanUp',
        context: cleanUpContext,
        stopOnError: false
      });
    } catch (error) {
      this.logger.error("[%s] Something goes wrong: %s\n%s", __pretty_name__, error.message, error.stack);
      this.properties.processExitCode = 1;
    }

    this.currentState = SimpleTestLauncher.States.ShuttingDown;

    this._doShutDown();
  }

  _doShutDown() {
    process.exit(this.properties.processExitCode);
  }

  async _getConfigurationFromEnvVars() {
    const configureEnvOptionsSchema = Joi.object().keys({
      project: Joi.object().keys({
        rootDirectoryPath: Joi.string()
      }).unknown(),
      reports: Joi.object().keys({
        rootDirectoryPath: Joi.string()
      }).unknown(),
      coverage: Joi.object().keys({
        enabled: Joi.boolean(),
        istanbul: Joi.object().keys({
          excludes: Joi.array().items(Joi.string()),
          nativeConfiguration: Joi.object().unknown(),
          reportDirectory: Joi.string()
        }).unknown(),
        summary: Joi.object().keys({
          enabled: Joi.boolean(),
          template: Joi.string()
        }).unknown()
      }).unknown(),
      test: Joi.object().keys({
        type: Joi.string(),
        specDirPath: Joi.string().allow(null),
        jasmine: Joi.object().keys({
          features: Joi.object().keys({
            enableJunitXml: Joi.boolean(),
            enableHtml: Joi.boolean()
          }).unknown(),
          nativeConfig: Joi.object().unknown()
        }).unknown()
      }).unknown()
    }).unknown().required();

    const inputConfiguration = {
      project: {
        rootDirectoryPath: _.get(process.env, 'ZE_CONFIG_PROJECT_ROOT_DIR_PATH', undefined)
      },
      reports: {
        rootDirectoryPath: _.get(process.env, 'ZE_CONFIG_REPORTS_ROOT_DIR_PATH', undefined)
      },
      coverage: {
        enabled: _.get(process.env, 'ZE_CONFIG_COVERAGE_ENABLED', undefined),
        istanbul: {
          excludes: _.filter(_.map(_.split(_.get(process.env, 'ZE_CONFIG_COVERAGE_EXCLUDES', ''), ','), item => item.trim()), item => item.length > 0),
          reportDirectory: _.get(process.env, 'ZE_CONFIG_COVERAGE_REPORT_DIR_NAME', undefined)
        },
        summary: {
          enabled: _.get(process.env, 'ZE_CONFIG_COVERAGE_SUMMARY_ENABLED', undefined),
          template: _.get(process.env, 'ZE_CONFIG_COVERAGE_SUMMARY_TEMPLATE', undefined)
        }
      },
      test: {
        type: _.get(process.env, 'ZE_CONFIG_TEST_TYPE', undefined),
        specDirPath: _.get(process.env, 'ZE_CONFIG_TEST_SPEC_DIR_PATH', undefined),
        jasmine: {
          features: {
            enableJunitXml: _.get(process.env, 'ZE_CONFIG_TEST_ENABLE_REPORT_JUNIT', undefined),
            enableHtml: _.get(process.env, 'ZE_CONFIG_TEST_ENABLE_REPORT_HTML', undefined)
          }
        }
      }
    };

    const {error, value: validatedConfiguration} = Joi.validate(inputConfiguration, configureEnvOptionsSchema);
    if (error) {
      throw error;
    }

    // Create the new object with the value parsed
    const configuration = {
      project: {
        rootDirectoryPath: _.get(validatedConfiguration, 'project.rootDirectoryPath', undefined)
      },
      reports: {
        rootDirectoryPath: _.get(validatedConfiguration, 'reports.rootDirectoryPath', undefined)
      },
      coverage: {
        enabled: _.get(validatedConfiguration, 'coverage.enabled', undefined),
        istanbul: {
          excludes: _.get(validatedConfiguration, 'coverage.istanbul.excludes', undefined),
          reportDirectory: _.get(validatedConfiguration, 'coverage.istanbul.reportDirectory', undefined)
        },
        summary: {
          enabled: _.get(validatedConfiguration, 'coverage.summary.enabled', undefined),
          template: _.get(validatedConfiguration, 'coverage.summary.template', undefined)
        }
      },
      test: {
        type: _.get(validatedConfiguration, 'test.type', undefined),
        specDirPath: _.get(validatedConfiguration, 'test.specDirPath', undefined),
        jasmine: {
          features: {
            reports: {
              enableJunitXml: _.get(validatedConfiguration, 'test.jasmine.features.enableJunitXml', undefined),
              enableHtml: _.get(validatedConfiguration, 'test.jasmine.features.enableHtml', undefined)
            }
          }
        }
      }
    };

    return {
      configuration: configuration,
      rawConfiguration: inputConfiguration
    };
  }

  async _getUserConfigurationFromObject(userOptions) {
    const configurationStepSchema = () => Joi.object().keys({
      handler: Joi.func().required()
    }).unknown();

    const configureOptionsSchema = Joi.object().keys({
      project: Joi.object().keys({
        rootDirectoryPath: Joi.string().required()
      }).unknown().required(),
      reports: Joi.object().keys({
        rootDirectoryPath: Joi.string()
      }).unknown(),
      coverage: Joi.object().keys({
        enabled: Joi.boolean(),
        istanbul: Joi.object().keys({
          excludes: Joi.array().items(Joi.string()),
          nativeConfiguration: Joi.object().unknown(),
          reportDirectory: Joi.string()
        }).unknown(),
        summary: Joi.object().keys({
          enabled: Joi.boolean(),
          template: Joi.string()
        }).unknown()
      }).unknown().required(),
      test: Joi.object().keys({
        type: Joi.string().required(),
        specDirPath: Joi.string().allow(null),
        specFileFilters: Joi.array().items(
          Joi.string(),
          Joi.func(),
          Joi.object().type(RegExp)
        ),
        jasmine: Joi.object().keys({
          features: Joi.object().keys({
            enableJunitXml: Joi.boolean(),
            enableHtml: Joi.boolean()
          }).unknown(),
          nativeConfig: Joi.object().unknown()
        }).unknown()
      }).unknown().required(),
      steps: Joi.object().keys({
        preLaunch: Joi.array().items(configurationStepSchema()),
        preTestLaunch: Joi.array().items(configurationStepSchema()),
        postExecution: Joi.array().items(configurationStepSchema()),
        cleanUp: Joi.array().items(configurationStepSchema())
      }).unknown()
    }).unknown().required();


    const {error, value: validatedOptions} = Joi.validate(userOptions, configureOptionsSchema);
    if (error) {
      throw error;
    }

    const configuration = {
      // From the input
      project: {
        rootDirectoryPath: _.get(validatedOptions, 'project.rootDirectoryPath')
      },
      reports: {
        rootDirectoryPath: _.get(validatedOptions, 'reports.rootDirectoryPath')
      },
      test: {
        type: _.get(validatedOptions, 'test.type'),
        specDirPath: _.get(validatedOptions, 'test.specDirPath'),
        specFileFilters: _.get(userOptions, 'test.specFileFilters'),
        jasmine: {
          features: {
            reports: {
              enableJunitXml: _.get(validatedOptions, 'test.jasmine.features.enableJunitXml'),
              enableHtml: _.get(validatedOptions, 'test.jasmine.features.enableHtml')
            }
          }
        }
      },
      coverage: {
        istanbul: {
          excludes: _.get(validatedOptions, 'coverage.istanbul.excludes'),
          nativeConfiguration: _.get(userOptions, 'coverage.istanbul.nativeConfiguration'),
          reportDirectory: _.get(validatedOptions, 'coverage.istanbul.reportDirectory')
        },
        summary: {
          enabled: _.get(validatedOptions, 'coverage.summary.enabled'),
          template: _.get(userOptions, 'coverage.summary.template')
        }
      },
      steps: _.assign({}, userOptions.steps)
    };

    return {
      rawConfiguration: userOptions,
      configuration: configuration
    };
  }

  async configure(options) {
    // const __pretty_name__ = "configure";
    const configs = {
      fromEnv: undefined,
      fromUsr: undefined
    };

    configs.fromUsr = await this._getUserConfigurationFromObject(options);
    configs.fromEnv = await this._getConfigurationFromEnvVars();

    this.properties.rawInputConfiguration.configurationFunc = configs.fromUsr.rawConfiguration;
    this.properties.rawInputConfiguration.fromEnvVariables = configs.fromEnv.rawConfiguration;

    const configuration = _.mergeWith({}, this.properties.configuration, configs.fromUsr.configuration, configs.fromEnv.configuration, lodashMergerWithIgnoreNil);
    // For path to be absolute
    configuration.project.rootDirectoryPath = pathUtil.resolve(configuration.project.rootDirectoryPath);
    if (configuration.reports.rootDirectoryPath) {
      configuration.reports.rootDirectoryPath = pathUtil.resolve(configuration.reports.rootDirectoryPath);
    }

    this.properties.configuration = configuration;

    // Those are the one use
    this.properties.steps.preLaunch = _.concat([], this.properties.steps.preLaunch, configs.fromUsr.rawConfiguration.steps.preLaunch);
    this.properties.steps.preTestLaunch = _.concat([], this.properties.steps.preTestLaunch, configs.fromUsr.rawConfiguration.steps.preTestLaunch);
    this.properties.steps.postExecution = _.concat([], this.properties.steps.postExecution, configs.fromUsr.rawConfiguration.steps.postExecution);
    this.properties.steps.cleanUp = _.concat([], this.properties.steps.cleanUp, configs.fromUsr.rawConfiguration.steps.cleanUp);

    this.currentState = SimpleTestLauncher.States.Ready;
  }

  async _execSteps(options) {
    if (!_.isArray(this.properties.steps[options.groupName])) {
      throw new Error(`Cannot find the steps for '${options.groupName}'`);
    }

    // Create a new step list which contains internal steps and external
    const internalSteps = [];
    const prefixName = '_autoStep' + options.groupName.substr(0, 1).toUpperCase() + options.groupName.substr(1);
    const functionNames = _.sortBy(_.filter(classUtils.getInstanceMethodNames(this), test => {
      return _.startsWith(test, prefixName);
    }));

    _.each(functionNames, functionName => {
      if (_.isFunction(this[functionName])) {
        internalSteps.push({
          handler: this[functionName].bind(this)
        });
      }
    });

    const steps = _.concat([], internalSteps, this.properties.steps[options.groupName]);
    if (steps.length === 0) {
      return;
    }

    const errors = [];

    for (let i = 0, l = steps.length; i < l; ++i) {
      try {
        const result = await steps[i].handler(options.context);
        if (_.isObjectLike(result) && result.stopWorkflow === true) {
          break;
        }
      } catch (error) {
        if (!options.stopOnError) {
          errors.push(error);
        } else {
          throw error;
        }
      }
    }

    if (errors.length > 0) {
      throw errors[0];
    }
  }

  _createStepBaseContext() {
    return {
      launcher: this,
      logger: this.logger,
      executionContext: this.properties.executionContext
    };
  }

  async createJasmineReporters() {
    const reporters = [];

    if (this.properties.configuration.test.jasmine.features.reports.enableJunitXml) {
      const jasmineReporters = require('jasmine-reporters'),
        reporter = new jasmineReporters.JUnitXmlReporter({
          savePath: this.getRootReportTestPath() + "/juint/xml",
          consolidateAll: false
        });

      reporters.push(reporter);
    }

    if (this.properties.configuration.test.jasmine.features.reports.enableHtml) {
      const Jasmine2HtmlReporter = require('protractor-jasmine2-html-reporter'),
        reporter = new Jasmine2HtmlReporter({
          savePath: this.getRootReportTestPath() + "/html",
          screenshotsFolder: 'images',
          takeScreenshots: false
        });

      reporters.push(reporter);
    }

    return {
      reporters: reporters
    };
  }

  async _filterTestSpecFiles(options) {
    const __pretty_name__ = '_filterTestSpecFiles';

    const promiseLimitUtils = require('promise-limit');

    const _filterTestSpecFilesOptionsSchema = Joi.object().keys({
      filePaths: Joi.array().items(Joi.string()).required()
    }).unknown().required();

    const {error, value: validatedOptions} = Joi.validate(options, _filterTestSpecFilesOptionsSchema);
    if (error) {
      throw error;
    }

    if (this.properties.configuration.test.specFileFilters.length === 0) {
      return {
        filePathsFiltered: options.filePaths
      };
    }

    // Generate the filters
    const filters = [];
    _.each(this.properties.configuration.test.specFileFilters, filter => {
      if (_.isString(filter) || _.isRegExp(filter)) {
        let regexpFilter = filter;
        if (_.isString(filter)) {
          // Create a regexp
          regexpFilter = new RegExp(filter);
        }

        filters.push(function (filePath) {
          return new Promise(resolve => {
            const matches = filePath.match(regexpFilter);
            const ignoreFile = _.isArray(matches) && matches.length > 0;
            return resolve({
              ignore: ignoreFile
            });
          });
        });
      } else {
        filters.push(filter);
      }
    });

    const filePathsFiltered = [];

    const handleFilePath = async filePath => {
      // Call all filters
      let keepFile = true;

      const handleFilter = async (filter, index) => {
        try {
          const result = await filter(filePath);

          if (result && result.ignore === true) {
            keepFile = false;
          }
        } catch (error) {
          this.logger.error("[%s] Filter %d failed: %s\n%s", __pretty_name__, index, error.message, error.stack);
          throw error;
        }
      };

      const limit = promiseLimitUtils(20);
      await Promise.all(
        _.map(filters, (filter, index) => limit(() => handleFilter(filter, index)))
      );

      if (keepFile) {
        filePathsFiltered.push(filePath);
      }
    };

    const limit = promiseLimitUtils(20);
    await Promise.all(
      _.map(validatedOptions.filePaths, filePath => limit(() => handleFilePath(filePath)))
    );

    return {
      filePathsFiltered: filePathsFiltered
    };
  }

  async _instantiateAndInitJasmine() {
    const __pretty_name__ = '_instantiateAndInitJasmine';

    const Jasmine = require('jasmine');
    const jasmineOptions = {
      projectBaseDir: this.getProjectRootDirPath()
    };

    const jasmineInstance = new Jasmine(jasmineOptions);
    const {configuration: jasmineConfig} = await this.fetchJasmineConfiguration();

    jasmineInstance.loadConfig(jasmineConfig);
    jasmineInstance.env.configure(jasmineConfig.env || {});

    const {reporters} = await this.createJasmineReporters();

    _.each(reporters, reporter => jasmineInstance.addReporter(reporter));
    jasmineInstance.env.executionContext = this.properties.executionContext;

    const specFilesResponse = await this._filterTestSpecFiles({
      filePaths: jasmineInstance.specFiles
    });

    jasmineInstance.specFiles = specFilesResponse.filePathsFiltered || [];
    this.logger.debug("[%s] Use those files:\n%s\n", __pretty_name__, _.join(specFilesResponse.filePathsFiltered, "\n"));
    this.properties.jasmine.instance = jasmineInstance;
  }

  async _autoStepPostExecutionSaveCoverage(context) {
    if (!this.isCodeCoverageEnabled()) {
      return;
    }

    const istanbul = require('istanbul'),
      path = require('path'),
      fse = require('fs-extra'),
      config = context.launcher.properties.istanbul.config,
      cov = global[context.launcher.properties.istanbul.coverageVariableName],
      reportingDirectory = path.resolve(config.reporting.dir()),
      coverageResultFilePath = path.resolve(reportingDirectory, 'coverage.json'),
      reporter = new istanbul.Reporter(config);

    // Configure the reporter
    reporter.dir = reportingDirectory;
    reporter.addAll(config.reporting.reports());
    // Could also add manual reporters:
    // reporter.add('text');
    // reporter.add('text');
    // reporter.add('text-summary');
    // reporter.add('text-summary');

    await fse.ensureDir(reportingDirectory);

    const writeReports = () => {
      return new Promise((resolve, reject) => {
        const collector = new istanbul.Collector();
        collector.add(cov);

        reporter.write(collector, true, function (error) {
          if (error) {
            reject(error);
          } else {
            resolve();
          }
        });
      });
    };

    await writeReports();
    await fse.writeJson(coverageResultFilePath, cov);

    if (_.get(this.properties.configuration, 'coverage.summary.enabled') !== true) {
      return;
    }

    // We do the summary
    const collector = new istanbul.Collector();
    collector.add(cov);

    const summaries = [];
    collector.files().forEach(file => {
      summaries.push(istanbul.utils.summarizeFileCoverage(collector.fileCoverageFor(file)));
    });

    const knownKeys = ['lines', 'statements', 'functions', 'branches'];

    const doSum = property => {
      let total = 0;
      _.each(knownKeys, key => {
        total += finalSummary[key][property];
      });
      return total;
    };

    const finalSummary = istanbul.utils.mergeSummaryObjects.apply(null, summaries);
    const templateVariables = {
      total: {
        percent: Number((doSum('pct') / 4).toFixed(2)),
        pct: doSum('pct') / 4,
        total: doSum('total'),
        covered: doSum('covered'),
        skipped: doSum('skipped')
      }
    };
    _.each(knownKeys, key => {
      templateVariables[key] = _.cloneDeep(finalSummary[key]);
      templateVariables[key].percent = templateVariables[key].pct;
    });

    const processTemplate = require('./utils/process-template').processTemplate;

    const {data} = await processTemplate({
      template: _.get(this.properties.configuration, 'coverage.summary.template'),
      variables: templateVariables
    });
    process.stdout.write(`\n\n${data}\n\n`);
  }

  async start() {
    const __pretty_name__ = "start";

    if (this.isRunning()) {
      this.logger.info("Already running");
      return;
    } else if (!this.isReady()) {
      this.logger.error("Launcher not ready");
      throw new Error("Launcher not ready");
    }

    /*
      1 - We do a jasmine init
      2 - We do a prelaunch step
      3 - We start coverage if necessary
      4 - We do the pre test launch
      4 - We launch jasmine
     */
    try {
      await this._instantiateAndInitJasmine();
      const preLaunchContext = this._createStepBaseContext();

      await this._execSteps({
        groupName: 'preLaunch',
        context: preLaunchContext
      });

      if (this.isCodeCoverageEnabled()) {
        await this._startCoverage();
      }

      const preTestLaunchContext = this._createStepBaseContext();

      await this._execSteps({
        groupName: 'preTestLaunch',
        context: preTestLaunchContext
      });

      await this._startTest();
      this.currentState = SimpleTestLauncher.States.Running;
    } catch (error) {
      this.logger.error("[%s][Failed] %s\n%s", __pretty_name__, error.message, error.stack);
      const cleanUpContext = this._createStepBaseContext();
      try {
        await this._execSteps({
          groupName: 'cleanUp',
          context: cleanUpContext,
          stopOnError: false
        });
      } catch (e) {
        this.logger.error("Error while running clean up: %s\n%s", e.message, e.stack);
      }

      throw error;
    }
  }

  async run() {
    const __pretty_name__ = "run";

    try {
      await this.start();
    } catch (error) {
      this.logger.error("[%s] Failed to start : %s\n%s", __pretty_name__, error.message, error.stack);
      this.properties.processExitCode = 1;
      await this._doShutDown();
    }
  }
}

exports.SimpleTestLauncher = module.exports.SimpleTestLauncher = SimpleTestLauncher;

